#ifndef NO_RCSID
const char* QN_Logger_rcsid =
    "$Header: /u/drspeech/repos/quicknet2/QN_Logger.cc,v 1.6 2006/01/10 02:59:09 davidj Exp $";
#endif

// User interaction routines.

/* Must include the config.h file first */
#include <QN_config.h>
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "QN_types.h"
#include "QN_Logger.h"

#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS (0)
#define EXIT_FAILURE (1)
#endif

////////////////////////////////////////////////////////////////
// Friendly interfaces to the va_list based logging functions.

void
QN_Logger::error(const char* objname, const char* format, ...)
{
    va_list va;

    va_start(va, format);
    verror(objname, format, va);
    va_end(va);
}

void
QN_Logger::warning(const char* objname, const char* format, ...)
{
    va_list va;

    va_start(va, format);
    vwarning(objname, format, va);
    va_end(va);
}

void
QN_Logger::log(const char* objname, const char* format, ...)
{
    va_list va;

    va_start(va, format);
    vlog(objname, format, va);
    va_end(va);
}

void
QN_Logger::output(const char* format, ...)
{
    va_list va;

    va_start(va, format);
    vlog(NULL, format, va);
    va_end(va);
}

////////////////////////////////////////////////////////////////

QN_ClassLogger::QN_ClassLogger(int a_debug, const char* a_objname,
			       const char* a_instname)
    : debug(a_debug),
      name(new char[strlen(a_objname)+strlen(a_instname) + 2])
{
    strcpy(name, a_objname);
    strcat(name, ".");
    strcat(name, a_instname);
}

QN_ClassLogger::~QN_ClassLogger()
{
    delete[] name;
}

void
QN_ClassLogger::output(const char* format, ...) const
{
    va_list va;
    
    va_start(va, format);
    QN_logger->vlog(NULL, format, va);
    va_end(va);
}

void
QN_ClassLogger::log(int level, const char* format, ...) const
{
    va_list va;
    
    va_start(va, format);
    if (debug>=level)
    {
	QN_logger->vlog(name, format, va);
    }
    va_end(va);
}

void
QN_ClassLogger::error(const char* format, ...) const
{
    va_list va;
    
    va_start(va, format);
    QN_logger->verror(name, format, va);
    va_end(va);
}

void
QN_ClassLogger::warning(const char* format, ...) const
{
    va_list va;
    
    va_start(va, format);
    QN_logger->vwarning(name, format, va);
    va_end(va);
}


////////////////////////////////////////////////////////////////

QN_FileToName::QN_FileToName()
{
    size_t i;

    for (i=0; i<MAX_ENTRIES; i++)
	fp[i] = NULL;
}

QN_FileToName::~QN_FileToName()
{
    size_t i;

    for (i=0; i<MAX_ENTRIES; i++)
    {
	if (fp[i] !=NULL)
	{
	    assert(filename[i]!=NULL);
// Cannot use this warning as most error exits don't close files.	    
//	    QN_WARN("QN_FileToName", "file %s not closed before exit.",
//		    filename[i]);
	    delete[] filename[i];
	}
    }
}

void
QN_FileToName::create_entry(FILE* afp, const char* afilename)
{
    size_t i;

    if (afp==stdin || afp==stdout || afp==stderr)
	return;
    for (i=0; i<MAX_ENTRIES; i++)
    {
	if (fp[i]==NULL)
	{
	    fp[i] = afp;
	    size_t len = strlen(afilename);
	    char* buf = new char[len];
	    strncpy(buf, afilename, len);
	    filename[i] = buf;
	    return;
	}
    }
    QN_ERROR("QN_FileToName", "failed to find space for new filename in create_entry.");
}

void
QN_FileToName::delete_entry(FILE* afp)
{
    size_t i;

    if (afp==stdin || afp==stdout || afp==stderr)
	return;
    for (i=0; i<MAX_ENTRIES; i++)
    {
	if (fp[i]==afp)
	{
	    fp[i] = NULL;
	    delete [] filename[i];
	    filename[i] = NULL;
	    return;
	}
    }
    QN_ERROR("QN_FileToName", "failed to find file in delete_entry.");
}

const char*
QN_FileToName::lookup(FILE* afp)
{
    size_t i;

    if (afp==stdin)
	return("(stdin)");
    else if (afp==stdout)
	return("(stdout)");
    else if (afp==stderr)
	return("(stderr)");
    else
    {
	for (i=0; i<MAX_ENTRIES; i++)
	{
	    if (fp[i]==afp)
	    {
		return(filename[i]);
	    }
	}
    }
    // Only here if we cannot find an entry.
    // Note that this is not necessarily an error as programs using the
    // library may have used "fopen" rather than "QN_open".
    return("(unknown)");
}

// The actuall FileToName object.

QN_FileToName QN_filetoname;
