// -*- C++ -*-
// $Header: /u/drspeech/repos/quicknet2/QN_MLPWeightFile_MLPW.cc,v 1.10 2006/01/05 20:37:44 davidj Exp $
// 
// QN_MLPWeightFle_MLPW.cc
//
// Define a new *binary* MLP Weight file format
// 1997jun19 dpwe@icsi.berkeley.edu
//

#ifndef NO_RCSID
const char* QN_MLPWeightFile_MLPW_rcsid =
    "$Header: /u/drspeech/repos/quicknet2/QN_MLPWeightFile_MLPW.cc,v 1.10 2006/01/05 20:37:44 davidj Exp $";
#endif

/* Must include the config.h file first */
#include <QN_config.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <math.h>


#include "QN_types.h"
#include "QN_Logger.h"
#include "QN_utils.h"
#include "QN_libc.h"
#include "QN_intvec.h"
#include "QN_fltvec.h"
#include "QN_MLPWeightFile_MLPW.h"

#define QN_MLP3_HEADER QN_WEIGHTS_NONE

#ifndef min
#define min(a,b) ((a)<(b)?(a):(b))
#endif

#ifndef max
#define max(a,b) ((a)>(b)?(a):(b))
#endif

// Name of this class
const char* QN_MLPWeightFile_MLPW::CLASSNAME = "QN_MLPWeightFile_MLPW";

int QN_MLPWeightFile_MLPW::read_header(void) {
    // Read the header of an MLPW file
    char buf[QN_MLPW_SIZEOFMAGIC];
    int nr;
    QNInt32 tmp;
    int version;

    if ((nr = fread(buf, QN_MLPW_SIZEOFMAGIC, 1, stream))) {
	if (strncmp(buf, QN_MLPW_MAGIC, QN_MLPW_SIZEOFMAGIC) == 0) {
	    // Magic number looks good - read other fields
	    nr += fread(&tmp, sizeof(QNInt32), 1, stream);
	    version = qn_btoh_i32_i32(tmp);
	    if (version == QN_MLPW_VERSION) {
		nr += fread(&tmp, sizeof(QNInt32), 1, stream);
		netType = qn_btoh_i32_i32(tmp);
		nr += fread(&tmp, sizeof(QNInt32), 1, stream);
		nLayers = qn_btoh_i32_i32(tmp);
		nr += fread(&tmp, sizeof(QNInt32), 1, stream);
		nSections = qn_btoh_i32_i32(tmp);
		nr += fread(&tmp, sizeof(QNInt32), 1, stream);
		dtype = static_cast<QN_MLPW_dtype>(qn_btoh_i32_i32(tmp));

		int i;
		// Read layer sizes
		if (nLayers < QN_MLPW_MAX_NLAYERS) {
		    for (i = 0; i < nLayers; ++i) {
			nr += fread(&tmp, sizeof(QNInt32), 1, stream);
			layerSize[i] = qn_btoh_i32_i32(tmp);
		    }
		}
		// Check that we read enough
		if (nr == 6+nLayers) {
		    // success
		    return 1;
		}
	    }
	}
    }
    // fail
    return 0;    
}

int QN_MLPWeightFile_MLPW::write_header(void) {
    // Write the header of an MLPW file
    int nw;
    QNInt32 tmp;

    nw = fwrite(QN_MLPW_MAGIC, QN_MLPW_SIZEOFMAGIC, 1, stream);
    tmp = qn_htob_i32_i32(QN_MLPW_VERSION);
    nw += fwrite(&tmp, sizeof(QNInt32), 1, stream);
    tmp = qn_htob_i32_i32(netType);
    nw += fwrite(&tmp, sizeof(QNInt32), 1, stream);
    tmp = qn_htob_i32_i32(nLayers);
    nw += fwrite(&tmp, sizeof(QNInt32), 1, stream);
    tmp = qn_htob_i32_i32(nSections);
    nw += fwrite(&tmp, sizeof(QNInt32), 1, stream);
    tmp = qn_htob_i32_i32(dtype);
    nw += fwrite(&tmp, sizeof(QNInt32), 1, stream);
    int i;
    for (i = 0; i < nLayers; ++i) {
	tmp = qn_htob_i32_i32(layerSize[i]);
	nw += fwrite(&tmp, sizeof(QNInt32), 1, stream);
    }
    // Check that we wrote enough
    if (nw == 6+nLayers) {
	// success
	return 1;
    }
    // fail
    return 0;    
}

int QN_MLPWeightFile_MLPW::read_sect_header(int sect, int numvals)
{
    int nr;
    QNUInt32 tmp;

    if ((nr = fread(&tmp, sizeof(QNUInt32), 1, stream))) {
	tmp = qn_btoh_i32_i32(tmp);
	if ((QN_SectionSelector)tmp == sect) {
	    // It looks like the right section
	    tmp = 0;
	    nr += fread(&tmp, sizeof(QNUInt32), 1, stream);
	    tmp = qn_btoh_i32_i32(tmp);
	    if (static_cast<QNUInt32>(numvals) != tmp) {
		QN_ERROR(CLASSNAME, "error reading MLPW sect header from \'%s\': expecting %d vals, got %d", 
			 filename, numvals, tmp);
	    }
	    // Check that we read enough
	    if (nr == 2) {
		// success
		return 1;
	    }
	}
    }
    // fail
    return 0;    
}

int QN_MLPWeightFile_MLPW::write_sect_header(int sect, 
					     int numvals)
{
    int nw;
    QNUInt32 tmp;

    // write section ID
    tmp = qn_htob_i32_i32((QNUInt32)sect);
    nw = fwrite(&tmp, sizeof(QNUInt32), 1, stream);
    // write vector size
    tmp = qn_htob_i32_i32((QNUInt32)numvals);
    nw += fwrite(&tmp, sizeof(QNUInt32), 1, stream);
    // Check that we wrote enough
    if (nw == 2) {
	// success
	return 1;
    }
    // fail
    return 0;    
}

// Constructor for an input weight file

QN_MLPWeightFile_MLPW::QN_MLPWeightFile_MLPW(int dbg, FILE* a_stream, 
					     QN_FileMode a_mode,
					     const char* a_name,
					     size_t n_layers, 
					     size_t layer_sizes[], 
					     QN_MLPW_ntype nt /* = none */,
					     QN_MLPW_dtype dt)
  : debug(dbg),
    mode(a_mode),
    stream(a_stream),
    filename(a_name)
{

    int i;
    int ec;

    if (a_mode==QN_READ)
    {
	if (debug>=2)
	{
	    QN_LOG(CLASSNAME, "Accessing MLPW file %s for reading.",
		   filename);
	}
	if (read_header() == 0) {
	    QN_ERROR(CLASSNAME, "error reading MLPW header from \'%s\'.", 
		     filename);
	}
	// Reading header sets up local fields
	if (n_layers > 0 && nLayers != n_layers) {
		QN_ERROR(CLASSNAME, "number of layers in MLPW file \'%s\' is"
			 " %lu - it should be %lu.", filename, nLayers, 
			 n_layers);
	}
	int i; 
	for (i = 0; i < n_layers; ++i) {
	    if (layerSize[i] != layer_sizes[i]) {
		QN_ERROR(CLASSNAME, "number of units in layer %i of MLPW "
			 "file \'%s\' is %lu - it should be %lu.", 
			 i, filename, layerSize[i], layer_sizes[i]);
	    }
	}

	// Read past first vector header, so we are ready to read data
	io_count = layerSize[0] * layerSize[1];
	io_state = QN_MLP3_INPUT2HIDDEN;
	read_sect_header(io_state, io_count);
    }
    else if (a_mode == QN_WRITE)
    {
	if (debug>=2)
	{
	    QN_LOG(CLASSNAME, "Accessing MLPW file %s for writing.",
		   filename);
	}
	dtype = dt;
	nLayers = n_layers;
	nSections = (nLayers-1)*2;

	int i;
	for (i = 0; i < nLayers; ++i) {
	    layerSize[i] = layer_sizes[i];
	}

	// io_state holds the QN_SectionSelector we're about to write,
	// Use a dummy value to say we haven't written the header yet
	
	io_state = QN_MLP3_HEADER;

    }
    else
	assert(0);

}

enum QN_SectionSelector
QN_MLPWeightFile_MLPW::get_weighttype(int section)
{
    switch(section)
    {
    case 0:
	return QN_MLP3_INPUT2HIDDEN;
	break;
    case 1:
	return QN_MLP3_HIDDEN2OUTPUT;
	break;
    case 2:
	return QN_MLP3_HIDDENBIAS;
	break;
    case 3:
	return QN_MLP3_OUTPUTBIAS;
	break;
    default:
	return QN_WEIGHTS_UNKNOWN;
	break;
    }
}

    
// Read in a buffer full of weights
size_t QN_MLPWeightFile_MLPW::read(float* buf, size_t count) {
#define QN_MLPW_BUF_SIZE 1024
    float *mybuf = new float[QN_MLPW_BUF_SIZE];

    assert(mode == QN_READ);

    // This is a hack so we can keep attempting to write past the end of file
    // without crashing
    if (io_count==0 || count==0)
	return(0);

    int thiscount, thiscountrem, thisblock;
    int bytespersamp = QN_MLPW_dt_sizeof(dtype);
    int wtsperbuf = QN_MLPW_BUF_SIZE*(sizeof(float)/max(sizeof(float),bytespersamp));
    int countrem = count;

    while (countrem > 0 & io_count > 0) {
	thiscount = min(countrem, io_count);
	thiscountrem = thiscount;
	while (thiscountrem > 0) {
	    thisblock = min(wtsperbuf, thiscountrem); 
	    // Do the read
	    if (fread(mybuf, bytespersamp, thisblock, stream) != thisblock) {
		QN_ERROR(CLASSNAME, "error reading weight file \'%s\' - %s",
			 filename, strerror(errno));
	    }
	    
	    // Convert data
	    switch (dtype) {
	    case QN_MLPW_dt_float:
		qn_btoh_vf_vf(thisblock, mybuf, buf);
		break;
	    case QN_MLPW_dt_double:
		qn_btoh_vd_vd(thisblock, (double *)mybuf, (double *)mybuf);
		qn_conv_vd_vf(thisblock, (double *)mybuf, buf);
		break;
	    default:
		QN_ERROR(CLASSNAME, "unknown MLPW data type %d", dtype);
	    }
	    buf += thisblock;

	    thiscountrem -= thisblock;
	}
	countrem -= thiscount;
	io_count -= thiscount;

	// Maybe update state machine
	if (io_count==0) {
	    int ix = 0;
	    switch(io_state) {
	    case QN_MLP3_INPUT2HIDDEN:
		io_state = QN_MLP3_HIDDEN2OUTPUT;
		io_count = layerSize[1] * layerSize[2];
		ix = 1;
		break;
	    case QN_MLP3_HIDDEN2OUTPUT:
		io_state = QN_MLP3_HIDDENBIAS;
		io_count = layerSize[1];
		ix = 2;
		break;
	    case QN_MLP3_HIDDENBIAS:
		io_state = QN_MLP3_OUTPUTBIAS;
		io_count = layerSize[2];
		ix = 3;
		break;
	    case QN_MLP3_OUTPUTBIAS:
		if (countrem > 0) {
		    QN_ERROR(CLASSNAME, "%d too many weights writing MLPW", 
			     countrem);
		}
		break;
	    default:
		assert(0);
	    }
	    read_sect_header(io_state, io_count);
	}
    }
    // We get here if we return as many elements as requested
    delete [] mybuf;

    return count;
}

// Write out a buffer full of weights
size_t QN_MLPWeightFile_MLPW::write(const float* buf, size_t count) {
    float *mybuf = new float[QN_MLPW_BUF_SIZE];

    assert(mode == QN_WRITE);

    // If we haven't already, write the header
    // deferred to here to give caller a chance to set up exponents
    // before they are written out.
    if (io_state == QN_MLP3_HEADER) {
	int ec;

	ec = write_header();
	if (ec<0) {
	    QN_ERROR(CLASSNAME, "error writing header, "
		     "weight file \'%s\' - %s.",
		     filename, strerror(errno));
	}
	io_count = layerSize[0] * layerSize[1];
	io_state = QN_MLP3_INPUT2HIDDEN;
	write_sect_header(io_state, io_count);
    }

    // This is a hack so we can keep attempting to write past the end of file
    // without crashing
    if (io_count==0 || count==0)
	return(0);

    int thiscount, thiscountrem, thisblock;
    int bytespersamp = QN_MLPW_dt_sizeof(dtype);
    int wtsperbuf = QN_MLPW_BUF_SIZE*(sizeof(float)/max(sizeof(float),bytespersamp));
    int countrem = count;

    while (countrem > 0 & io_count > 0) {
	thiscount = min(countrem, io_count);
	thiscountrem = thiscount;
	while (thiscountrem > 0) {
	    thisblock = min(wtsperbuf, thiscountrem); 
	    //qn_mul_vff_vf(thisblock, buf, invWeightScale, mybuf);
	    switch (dtype) {
	    case QN_MLPW_dt_float:
		qn_htob_vf_vf(thisblock, buf, mybuf);
		break;
	    case QN_MLPW_dt_double:
		qn_conv_vf_vd(thisblock, buf, (double *)mybuf);
		qn_htob_vd_vd(thisblock, (double *)mybuf, (double *)mybuf);
		break;
	    default:
		QN_ERROR(CLASSNAME, "unknown MLPW data type %d", dtype);
	    }

	    if (fwrite(mybuf, bytespersamp, thisblock, stream) != thisblock) {
		QN_ERROR(CLASSNAME, "error writing weight file \'%s\' - %s",
			 filename, strerror(errno));
	    }
	    buf += thisblock;
	    thiscountrem -= thisblock;
	}
	countrem -= thiscount;
	io_count -= thiscount;

	// Maybe update state machine
	if (io_count==0) {
	    int ix = 0;
	    switch(io_state) {
	    case QN_MLP3_INPUT2HIDDEN:
		io_state = QN_MLP3_HIDDEN2OUTPUT;
		ix = 1;
		io_count = layerSize[1] * layerSize[2];
		break;
	    case QN_MLP3_HIDDEN2OUTPUT:
		io_state = QN_MLP3_HIDDENBIAS;
		ix = 2;
		io_count = layerSize[1];
		break;
	    case QN_MLP3_HIDDENBIAS:
		io_state = QN_MLP3_OUTPUTBIAS;
		ix = 3;
		io_count = layerSize[2];
		break;
	    case QN_MLP3_OUTPUTBIAS:
		if (countrem > 0) {
		    QN_ERROR(CLASSNAME, "%d too many weights writing MLPW", 
			     countrem);
		}
		break;
	    default:
		assert(0);
	    }
	    write_sect_header(io_state, io_count);
	}
    }

    delete [] mybuf;

    return count;
}

enum QN_WeightMaj
QN_MLPWeightFile_MLPW::get_weightmaj(QN_SectionSelector)
{
    return QN_OUTPUTMAJOR;
}

enum QN_FileMode
QN_MLPWeightFile_MLPW::get_filemode()
{
    return mode;
}

